# -*- coding: utf-8 -*-
"""
Created on Thu Mar 08 15:40:08 2018

@author: han-luo
"""
from __future__ import  division
from collections import  namedtuple
import os, logging, subprocess, time, gc
import heapq, multiprocessing
import numpy as np

log = logging.getLogger(__name__)


"""
    BamProcess makes bam to bed file. formats as :
    
    column    :    Means
         0    :    Pair Name
         1    :    R1 mate Reference
         2    :    R1 mate Strand
         3    :    R1 mate Position
         4    :    R1 mate Length
         5    :    R1 mate AS score
         6    :    R1 mate Fragment Middle point
         7    :    R1 mate SNP Matching num(Non-Allelic results in 0)
         
         8    :    R2 mate Reference
         9    :    R2 mate Strand
        10    :    R2 mate Position
        11    :    R2 mate Length
        12    :    R2 mate AS score
        13    :    R2 mate Fragment Middle point
        14    :    R2 mate SNP Matching num(Non-Allelic results in 0)
        
        (candidate mate if it is possible)
        15    :    Candidate mate Reference
        16    :    Candidate mate Strand
        17    :    Candidate mate Position
        18    :    Candidate mate Length
        19    :    Candidate mate AS score
        20    :    Candidate mate Fragment Middle point
        21    :    Candidate mate SNP Matching num(Non-Allelic results in 0)
        22    :    Candidate Index for which mate.

"""



def uniqueIndex(data):
    """Returns a binary index of unique elements in an array data.
    This method is very memory efficient, much more than np.unique!
    It grabs only 9 extra bytes per record :)
    
    This Function comes from hiclib.(Author: Anton Goloborodko)
    Nice Function. :)
    """

    args = np.argsort(data)
    index = np.zeros(len(data), bool)
    myr = range(0, len(data), len(data) // 50 + 1) + [len(data)]
    for i in xrange(len(myr) - 1):
        start = myr[i]
        end = myr[i + 1]
        dataslice = data.take(args[start:end], axis=0)
        ind = dataslice[:-1] != dataslice[1:]
        index[args[start:end - 1]] = ind
        if end != len(data):
            if data[args[end - 1]] != data[args[end]]:
                index[args[end - 1]] = True
        else:
            index[args[end - 1]] = True

    return index

def Chunk_sort(chunk,OutPath):
    """
    Sorting the chunk bed for duplicate removing.
        sorted by :
            chr1  : chromosome of mate 1     
            flag1 : strand     of mate 1     
            pos1  : position   of mate 1     
                
            chr1  : chromosome of mate 2     
            flag2 : strand     of mate 2     
            pos2  : position   of mate 2
            
        sort order is from top to bottom.
    """
    tmp = []
    f = open(chunk,'r')
    for line in f:
        line = line.split()
        tmp.append(line)
    f.close()
    sorted_tmp = sorted(tmp,key = lambda x : (x[1],x[2],int(x[3]),x[8],x[9],int(x[10])))
        
    chunk_name = os.path.split(chunk)[-1]
    chunk_name = chunk_name.replace('.bed','_Sorted.bed')
    new_chunk = os.path.join(OutPath,chunk_name)
    with open(new_chunk,'w') as o:
        for line in sorted_tmp:
            o.writelines('\t'.join(line)+'\n')
        
        #subprocess.call(['rm',chunk])
        
    return new_chunk

def Merge_Sorted_Files(key, *iterables):
    """
        Merge the sorted chunks.
    """
    Keyed = namedtuple("Keyed",["key","obj"])
    if key is None:
        keyed_iterables = iterables
    else:
        keyed_iterables = [(Keyed(key(obj.split()), obj) for obj in iterable)
                            for iterable in iterables]
    for element in heapq.merge(*keyed_iterables):
        yield element.obj
        
        
        
# A Customized HiC Filtering Class
class cFiltering(object):
    
    def __init__(self, bedPath, Out_Path, threads, Collection = True,Allelic = 'NonAllelic'):
        
        
        #---------------Initialization the parameters--------
        self.bedPath = bedPath
        self.Allelic = Allelic
        self.Out_Path = Out_Path            #Non-Allelic Out Path.
        self.step = 20000000                #dups remove pairs ID number in memory. 
        self.Collection = Collection        #Whether rm the chunk bed files.
        self.Threads = threads              #Threads Number        
        
    def _StrAsciID(self,string):
        """
            return the Ascii ID sum of a string.
            
        """
        return sum(ord(i) for i in string)
    
    def _GetpairID(self,line):
        """
            return a pair of interaticon ID for duplicates filtering.
            
        """
        line = line.strip().split('\t')
        c1 = self._StrAsciID(line[1].lstrip('chr'))
        f1 = line[2];p1 = line[3]
        
        c2 = self._StrAsciID(line[8].lstrip('chr'))
        f2 = line[9];p2 = line[10]
        
        return int(str(c1)+f1+p1+str(c2)+f2+p2)    
    
    def _gcsleep(self):
        """
            sleep for a second, run garbage collector, sleep again.
            I dont konw if it makes sense, but it definitely does not hurts.
        """
        for _ in range(3):
            time.sleep(0.1)
        gc.collect()
        for _ in range(3):
            time.sleep(0.1)
     
    
    
    def _Getprefix(self):
        """
            return the prefix of file name.
        """
        for bed in os.listdir(self.bedPath):
            if 'chunk' in bed:
                prefix = bed.split('chunk')[0]
                return prefix
        
        return 'tmp_'
    
    def _Mergechunks(self):
        """
            Get chunks list and return the mergeing Linux command. 
        """
        if self.Allelic != 'NonAllelic':
            bed_files = [os.path.join(self.bedPath,i) for i in os.listdir(self.bedPath) \
                        if self.Allelic in i and 'chunk' in i]
        else:
            bed_files = [os.path.join(self.bedPath,i) for i in os.listdir(self.bedPath) \
                        if 'chunk' in i]
        return ['cat'] + sorted(bed_files)
    
    
    def _Collection(self):
        """
            Remove the chunks bed. Only HiC_filtered bed retain.
            Return the Remove Linux command
        """
        if self.Allelic != 'NonAllelic':
            bed_files = [os.path.join(self.bedPath,i) for i in os.listdir(self.bedPath)\
                        if 'chunk' in i and self.Allelic in i]
        else:
            bed_files = [os.path.join(self.bedPath,i) for i in os.listdir(self.bedPath) \
                        if 'chunk' in i]
        return ['rm'] + bed_files
    
    def _MergeSortedFiles(self,key=None,*iterables):
        """
        """
        Keyed = namedtuple('Keyed', ['key', 'obj'])
        if key is None:
            keyed_iterables = iterables
        else:
            keyed_iterables = [(Keyed(key(obj.split()), obj) for obj in iterable)
                                for iterable in iterables]
        for element in heapq.merge(*keyed_iterables):
            yield element.obj
        
    
    def _HiC_Sorting(self,key = None):
        """
            Sorting file for HiC duplicates filtering.
        """
        threads = self.Threads
        Allelic = self.Allelic
        if Allelic == 'NonAllelic':
            Allelic = 'chunk'
        bedPath = self.bedPath
        OutPath = self.Out_Path
        
        sort_Pool = multiprocessing.Pool(threads)
        
        chunks = [os.path.join(bedPath,i) for i in os.listdir(bedPath) if Allelic in i]
        log.log(21,'Chunk Sorting ...')
        #print 'Chunk Sorting ...'
        for fil in chunks:
            sort_Pool.apply_async(Chunk_sort,args=(fil,OutPath))
        
        sort_Pool.close()
        sort_Pool.join()
        log.log(21,'Merging Sorted Chunks...')
        #print 'Merging Sorted Chunks ...'
        sorted_chunks = [os.path.join(OutPath,i) for i in os.listdir(OutPath) \
                            if Allelic in i and 'Sorted' in i]
        chunks = []
        for fil in sorted_chunks:
            out = open(fil,'r')
            chunks.append(out)
        
        prefix = self._Getprefix()
        
        if Allelic == 'chunk':
            Total_Out = os.path.join(OutPath,prefix+'NonAllelic.bed')
        else:            
            Total_Out = os.path.join(OutPath,prefix+Allelic+'.bed')
        with open(Total_Out,'w') as Out:
            Out.writelines(Merge_Sorted_Files(lambda x : (x[1],x[2],int(x[3]),x[8],x[9],int(x[10])),*chunks))
        self.Total_Out = Total_Out
        for IO in chunks:
            IO.close()
        collect = ['rm'] + sorted_chunks
        subprocess.call(collect)
        log.log(21,'Sorting Done')
        #print 'Sorting Done'
        
        
        
        
    
    def _Redundant_kind(self,line):
        """
            Pair Tagging if pair is Redundant pair.
            
            Self-Circle:  SC
                Pair deriving from self-circularized ligation product.
                The two reads are mapped to the same restriction fragment and face
                in opposite directions.
                
               R1                            R1
            <----                            ---->              
            ==============    or    ==============
                     ---->          <----
                     R2                R2
            
            DanglingEnds:  DE
                There can be may causes of such products,ranging from low ligation efficiency
                to poor strptavidin specificity.
                Both tow reads are mapped to the same restriction fragment and face
                towards each other.
                
            R1                                  R1
            ---->                            <----  
            ==============    or    ==============
                     <----          ---->
                        R2          R2
            
            UnkownMechanism:  UM
                Unkown sources.
                Both two reads are mapped to the same restriction fragment and same strand.
            
            ---->                   <----
            ==============    or    ==============
                     ---->                   <----
            
            ExtraDanglingEnds:  ED
                The two reads of these pairs are mapped to the different restriction
                fragments but face towards each other and are separated by less than
                the library size (500bp) interval.Such read pairs may contain true 
                contacts, but are largely contaminated, so we also remove these paris.
                
                R1                              R1
                ---->                        <---- 
            =========|======== or   ========|=========
                      <----            ---->
                          R2           R2
        """
        
        line = line.split()
        C1 = line[1];C2 = line[8]                           #chromosome
        strand1 = int(line[2]);strand2 = int(line[9])       #strand
        pos1 = int(line[3]);pos2 = int(line[10])            #position
        Frag1 = int(line[6]);Frag2 = int(line[13])          #Fragments Middle
        
        if C1 != C2:
            return False
        elif Frag1 == Frag2:
            
            if pos1 < pos2:
                if strand1 == 0 and strand2 == 16:
                    return 'DE'
                elif strand1 == 16 and strand2 == 0:
                    return 'SC'
                else:
                    return 'UM'
            else:
                if strand1 == 0 and strand2 == 16:
                    return 'SC'
                elif strand1 == 16 and strand2 == 0:
                    return 'DE'
                else:
                    return 'UM'
        else:
            if abs(pos1 - pos2) <= 500:
                if pos1 < pos2 and strand1 == 0 and strand2 == 16:
                    return 'ED'
                elif pos1 > pos2 and strand1 == 16 and strand2 == 0:
                    return 'ED'
                else:
                    return False
            else:
                return False
    
    
    
    def HiC_Filtering(self):
        """
            HiC Filtering for duplicates, same Fragments reads.
        """
        log.log(21,'HiC Filtering Start for %s Interactions' % self.Allelic)
        #print 'HiC Filtering Start for %s Interations' % self.Allelic
        log.log(21,'Sorting and HiC filtering...')
        
        self._HiC_Sorting()
        log.log(21,'Filtering start ...')
        #print 'Filtering start ...'
        self.Duplicates = 0
        self.SelfCircle = 0
        self.DanglingEnds = 0
        self.UnknownMechiansm = 0
        self.ExtraDanglingEnds = 0
        self.ValidPairs = 0
        prefix = self._Getprefix()
        if self.Allelic != 'NonAllelic':
            self.Outbed = os.path.join(self.Out_Path,prefix+self.Allelic+'_Valid.bed')
            Out_bed = open(os.path.join(self.Out_Path,prefix+self.Allelic+'_Valid.bed'),'w')
        else:
            self.Outbed = os.path.join(self.Out_Path,prefix+'Valid.bed')
            Out_bed = open(os.path.join(self.Out_Path,prefix+'Valid.bed'),'w')
        mark = 0                        #Initial value
        count = 0                       #Step num for Unique
        f = open(self.Total_Out,'r')
        for line in f:
            if (count+1) % 40000000 == 0:
                log.log(21,'%d Pairs have been filtered',count+1)
                #print '%d Pairs have been filtered' % (count + 1)
            count += 1
            
            ID = self._GetpairID(line)
            if ID != mark:
                mark = ID
                PM = self._Redundant_kind(line)
                
                if PM == False:
                    self.ValidPairs += 1
                    Out_bed.writelines(line)
                elif PM == 'SC':
                    self.SelfCircle += 1
                elif PM == 'DE':
                    self.DanglingEnds += 1
                elif PM == 'UM':
                    self.UnknownMechiansm += 1
                elif PM == 'ED':
                    self.ExtraDanglingEnds += 1
            else:
                self.Duplicates += 1
        
        log.log(21,'%d pairs have been filtered',count)
        #print '%d paris have been filtered' % count
        
        Static_results = ['Total Unique Pairs : %d' % count,
                          'Duplicate Pairs : %d' % self.Duplicates,
                          'Valid Pairs  is %d' % self.ValidPairs,
                          'Self Circle Pairs is %d' % self.SelfCircle,
                          'DanglingEnds Pairs is %d' % self.DanglingEnds,
                          'UnkownMechanism Pairs is %d' % self.UnknownMechiansm,
                          'ExtraDanglingEnds Pairs is %d' % self.ExtraDanglingEnds]
        
        Static_results = '\n'*2 + '\n'.join(Static_results) + '\n'
        
        log.log(21,'HiC Statics Results : %s',Static_results)
        #print 'HiC Statics Results : %s' % Static_results
        log.log(21,'HiC filtering for %s Interations Done! \n',self.Allelic)
        
        f.close()
        subprocess.call(['rm',self.Total_Out])
        Out_bed.close()
        
        if self.Collection == True:
            subprocess.call(self._Collection())
        


# A Customized Allelic Filtering Class
class aFiltering(object):
    
    def __init__(self,Maternal_bed,Paternal_bed,Out_Path,save_ID = False):
        
        
        #---------------Initialization the parameters--------
        self.Maternal_bed = Maternal_bed                #Maternal Filtered bed
        self.Paternal_bed = Paternal_bed                #Paternal Filtered bed
        self.Out_Path = Out_Path                        #Out
        self.ChunkStep = 10000000                       #chunk Step for Name sorting.
        self.MAX_DIFF_SCORE = 18                         #Max diff score 
        self.save_ID = save_ID                          #Save the pair ID.
    
    
    def _BigFileSorting(self,bed):
        """ 
            Sorting the Interation bed file(.bed) by reads name created by bamProcess.
        
        """
        
        from itertools import islice, count
        from contextlib2 import ExitStack
        from heapq import merge
        
        log.log(21,'Sorting file  %s',bed)
        #print 'Sorting file %s' % bed
        log.log(21,'chunk file in step line: %d and sorting',self.ChunkStep)
        #print 'chunk file in step line %d' % self.ChunkStep
        chunk_names = []
        
        OutPath = self.Out_Path
        if not os.path.exists(OutPath):
            os.mkdir(OutPath)
            
        prefix = os.path.split(bed)[1].replace('.bed','')
        with open(bed) as input_fil:
            for chunk_number in count(1):
                sorted_chunk = sorted(islice(input_fil,self.ChunkStep))
                if not sorted_chunk:
                    break
                
                chunk_name = prefix+'_chunks_{}.chk'.format(chunk_number)
                chunk_names.append(chunk_name)
                with open(os.path.join(OutPath,chunk_name),'w') as chunk_fil:
                    chunk_fil.writelines(sorted_chunk)
                #log.log(21,'chunk %d completed',chunk_number)
                #print 'chunk %d completed' % chunk_number
        
        log.log(21,'Merge the sorted chunks')
        #print 'Merging...'
        Out = os.path.join(OutPath,prefix + '_sorted.bed')
        with ExitStack() as stack, open(Out,'w') as output_fil:
            files = [stack.enter_context(open(os.path.join(OutPath,chunk))) for chunk in chunk_names]
            output_fil.writelines(merge(*files))
        
        log.log(21,'temp files collection')
        chunk_files = [os.path.join(OutPath,chunk) for chunk in chunk_names]
        subprocess.call(['rm']+chunk_files)
        subprocess.call(['rm']+[bed])
        log.log(21,'Done!')
        #print 'Done!'
        
        return Out
    
    def _intToString(self,lst):
        """
            Transform the int type to string type in the list.
        """
        return [str(i) for i in lst]
    
    def _is_Available_Candidate(self,Info):
        """
            Whether the Candidate mate can be used to Allelic selecting.
            1.
            
            -------|-------   ---------------
            R1            C          R2
            
            2.
            ---------------   -----|---------
                   R1         C            R2
                   
            For 1 :
                Candidate(C) must have the same fragment mapping results as R1.
            
            For 2 :
                Candidate(C) must have the same fragment mapping results as R2.
        """
        mask = False
        candidate = Info[-1]
        if candidate == 'R1':
            C1 = Info[1]
            C2 = Info[15]
            Frag1 = int(Info[6])
            Frag2 = int(Info[20])
            
            if C1 == C2 and Frag1 == Frag2:
                mask = True
                return mask
        else:
            C1 = Info[8]
            C2 = Info[15]
            Frag1 = int(Info[13])
            Frag2 = int(Info[20])
            
            if C1 == C2 and Frag1 == Frag2:
                mask = True
                return mask
        
        return mask
        
    
    
    
    
    def _sub_search(self,M_C, M_pos, M_score, M_SNPs,
                         P_C, P_pos, P_score, P_SNPs):
        """
            Using Parameters(mentioned in Allelic_Filtering) to finish Allelic selecting.
            
            Parameters:
            ----------
            
            M_C : Maternal reference.
            
            M_pos : Maternal mapping position.
            
            M_score : Maternal mapping punish score.
            
            M_SNPs : Maternal SNP sites matching counts.
            
            ...
            
            Paternal Parameters are the same.
            
        """
        
        #Mapping To the Same Position
        if M_C == P_C and abs(M_pos - P_pos) <= 5:
            if M_SNPs > (2 * P_SNPs):
                mark = 'M'
            elif (2 * M_SNPs) < P_SNPs:
                mark = 'P'
            else:
                mark = 'N'
        
        #Mapping To the Diff Position
        else:
            if (M_score - P_score) >= self.MAX_DIFF_SCORE and M_SNPs >= 2 * P_SNPs:
                mark = 'M'
            elif (P_score - M_score) >= self.MAX_DIFF_SCORE and P_SNPs >= 2 * M_SNPs:
                mark = 'P'
            else:
                mark = 'N'
        
        return mark
        




    
    def _Both_Mapping_line_Process(self):
        """
            Using Allelic searching rules for a single mapping results.
            Pair mapped to  both two allelic genome succeed.
            
            Parameters
            ----------
                self
                
            Returns
            ----------
            mark1 + mark2 : str
                NN ---> Bi-Allelic
                NM ---> R2 Maternal
                NP ---> R2 Paternal
                MN ---> R1 Maternal
                PN ---> R1 Paternal
                MM ---> R1,R2 Maternal
                PP ---> R1,R2 Paternal
                MP,PM ---> Regroup
            
            lines : list
                Information without ['both','R1','R2'] for IO stream.
            
        """
        
        self.M_1_C = self.M_Info[1];self.M_2_C = self.M_Info[8]
        self.P_1_C = self.P_Info[1];self.P_2_C = self.P_Info[8]
        
        self.M_1_pos = int(self.M_Info[3]);self.M_2_pos = int(self.M_Info[10])
        self.P_1_pos = int(self.P_Info[3]);self.P_2_pos = int(self.P_Info[10])
        
        self.M_1_Frag = int(self.M_Info[6]);self.M_2_Frag = int(self.M_Info[13])
        self.P_1_Frag = int(self.P_Info[6]);self.P_2_Frag = int(self.P_Info[13])
        
        self.M_1_score = int(self.M_Info[5]);self.M_2_score = int(self.M_Info[12])
        self.P_1_score = int(self.P_Info[5]);self.P_2_score = int(self.P_Info[12])
        
        self.M_1_SNPs = int(self.M_Info[7]);self.M_2_SNPs = int(self.M_Info[14])
        self.P_1_SNPs = int(self.P_Info[7]);self.P_2_SNPs = int(self.P_Info[14])
        
        if len(self.M_Info) == len(self.P_Info) == 15:
            
            #For R1
            mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                     self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
            
            if mark1 == 'N' or mark1 == 'M':
                line1 = [self.M_1_C,self.M_1_Frag]
            else:
                line1 = [self.P_1_C,self.P_1_Frag]
                
            #For R2
            mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                     self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
            
            if mark2 == 'N' or mark2 == 'M':
                line2 = [self.M_2_C,self.M_2_Frag]
            else:
                line2 = [self.P_2_C,self.P_2_Frag]
            
            line = line1 + line2
            
        
        elif len(self.M_Info) > 15 and len(self.P_Info) == 15:            
            candidate_mark = self._is_Available_Candidate(self.M_Info)
            candidate = self.M_Info[-1]
            
            mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                     self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
            
            if mark1 == 'N' or mark1 == 'M':
                line1 = [self.M_1_C,self.M_1_Frag]
            else:
                line1 = [self.P_1_C,self.P_1_Frag]                         
            
            mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                     self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
            
            if mark2 == 'N' or mark2 == 'M':
                line2 = [self.M_2_C,self.M_2_Frag]
            else:
                line2 = [self.P_2_C,self.P_2_Frag]
                
                
            if candidate == 'R1' and candidate_mark == True:
                if mark1 == 'N':
                    self.M_1_C = self.M_Info[15]
                    self.M_1_pos = int(self.M_Info[17])
                    self.M_1_Frag = int(self.M_Info[20])
                    self.M_1_score = int(self.M_Info[19])
                    self.M_1_SNPs = int(self.M_Info[21])
                    mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                             self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
                    
                    if mark1 == 'M':
                        line1 = [self.M_1_C,self.M_1_Frag]
                    elif mark1 == 'P':
                        line1 = [self.P_1_C,self.P_1_Frag]
                    else:
                        pass
                        
                else:
                    pass
                
            elif candidate == 'R2' and candidate_mark == True:
                if mark2 == 'N':
                    self.M_2_C = self.M_Info[15]
                    self.M_2_pos = int(self.M_Info[17])
                    self.M_2_Frag = int(self.M_Info[20])
                    self.M_2_score = int(self.M_Info[19])
                    self.M_2_SNPs = int(self.M_Info[21])
                    mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                             self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
                    
                    if  mark2 == 'M':
                        line2 = [self.M_2_C,self.M_2_Frag]
                    elif mark2 == 'P':
                        line2 = [self.P_2_C,self.P_2_Frag]
                    else:
                        pass
                        
                else:
                    pass
            
            else:
                pass
            
            line = line1 + line2
            
        
        elif len(self.M_Info) == 15 and len(self.P_Info) >= 15:
            candidate_mark = self._is_Available_Candidate(self.P_Info)
            candidate = self.P_Info[-1]
            
            mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                     self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
            
            if mark1 == 'N' or mark1 == 'M':
                line1 = [self.M_1_C,self.M_1_Frag]
            else:
                line1 = [self.P_1_C,self.P_1_Frag]
                         
            mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                     self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
            
            if mark2 == 'N' or mark2 == 'M':
                line2 = [self.M_2_C,self.M_2_Frag]
            else:
                line2 = [self.P_2_C,self.P_2_Frag]
            
            
            if candidate == 'R1' and candidate_mark == True:
                if mark1 == 'N':
                    self.P_1_C = self.P_Info[15]
                    self.P_1_pos = int(self.P_Info[17])
                    self.P_1_Frag = int(self.P_Info[20])
                    self.P_1_score = int(self.P_Info[19])
                    self.P_1_SNPs = int(self.P_Info[21])
                    mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                             self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
                    
                    if mark1 == 'M':
                        line1 = [self.M_1_C,self.M_1_Frag]
                    elif mark1 == 'P':
                        line1 = [self.P_1_C,self.P_1_Frag]
                    else:
                        pass
                    
                else:
                    pass
            elif candidate == 'R2' and candidate_mark == True:
                if mark2 == 'N':
                    self.P_2_C = self.P_Info[15]
                    self.P_2_pos = int(self.P_Info[17])
                    self.P_2_Frag = int(self.P_Info[20])
                    self.P_2_score = int(self.P_Info[19])
                    self.P_2_SNPs = int(self.P_Info[21])
                    mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                             self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
                    
                    if mark2 == 'M':
                        line2 = [self.M_2_C,self.M_2_Frag]
                    elif mark2 == 'P':
                        line2 = [self.P_2_C,self.P_2_Frag]
                    else:
                        pass
                        
                else:
                    pass
            else:
                pass
            
            line = line1 + line2
        
        else:
            candidate_mark_M = self._is_Available_Candidate(self.M_Info)
            candidate_mark_P = self._is_Available_Candidate(self.P_Info)
            
            candidate = self.M_Info[-1]     #Both have candidate. And will for the same mate.
            
            mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                     self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
            
            
            if mark1 == 'N' or mark1 == 'M':
                line1 = [self.M_1_C,self.M_1_Frag]
            else:
                line1 = [self.P_1_C,self.P_1_Frag]

                         
            mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                     self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
            
            
            if mark2 == 'N' or mark2 == 'M':
                line2 = [self.M_2_C,self.M_2_Frag]
            else:
                line2 = [self.P_2_C,self.P_2_Frag]
                
            
            if candidate == 'R1' and mark1 == 'N':
                if candidate_mark_M == True:
                    self.M_1_C = self.M_Info[15]
                    self.M_1_pos = int(self.M_Info[17])
                    self.M_1_Frag = int(self.M_Info[20])
                    self.M_1_score = int(self.M_Info[19])
                    self.M_1_SNPs = int(self.M_Info[21])
                    
                if candidate_mark_P == True:
                    self.P_1_C = self.P_Info[15]
                    self.P_1_pos = int(self.P_Info[17])
                    self.P_1_Frag = int(self.P_Info[20])
                    self.P_1_score = int(self.P_Info[19])
                    self.P_1_SNPs = int(self.P_Info[21])
                                
                mark1 = self._sub_search(self.M_1_C,self.M_1_pos,self.M_1_score,self.M_1_SNPs,
                                         self.P_1_C,self.P_1_pos,self.P_1_score,self.P_1_SNPs)
                
                if mark1 == 'M':
                    line1 = [self.M_1_C,self.M_1_Frag]
                elif mark1 == 'P':
                    line1 = [self.P_1_C,self.P_1_Frag]
                else:
                    pass
            
            
            elif candidate == 'R2' and mark2 == 'N':
                if candidate_mark_M == True:
                    self.M_2_C = self.M_Info[15]
                    self.M_2_pos = int(self.M_Info[17])
                    self.M_2_Frag = int(self.M_Info[20])
                    self.M_2_score = int(self.M_Info[19])
                    self.M_2_SNPs = int(self.M_Info[21])
                    
                if candidate_mark_P == True:
                    self.P_2_C = self.P_Info[15]
                    self.P_2_pos = int(self.P_Info[17])
                    self.P_2_Frag = int(self.P_Info[20])
                    self.P_2_score = int(self.P_Info[19])
                    self.P_2_SNPs = int(self.P_Info[21])
                
                mark2 = self._sub_search(self.M_2_C,self.M_2_pos,self.M_2_score,self.M_2_SNPs,
                                         self.P_2_C,self.P_2_pos,self.P_2_score,self.P_2_SNPs)
    
                if mark2 == 'M':
                    line2 = [self.M_2_C,self.M_2_Frag]
                elif mark2 == 'P':
                    line2 = [self.P_2_C,self.P_2_Frag]
                else:
                    pass
            
            else:
                pass
            
            line = line1 + line2
        
        if self.save_ID == True:
            
            return mark1 + mark2, [self.M_Info[0]]+line
        
        else:
            return mark1 + mark2, line
    
    
    
    
    
    
    def _Specific_Mapping_line_Process(self,Info):
        """
            When pair Only mapped to one Allelic Genome.
            
            Complete the Allelic selected Process.
            
            Parameters
            ----------
            Info : list
                Mapping Info list.
                
            Return
            ---------
            mark : str
                both ---> both mates can be allelic selected.
                R1   ---> Only R1 can be allelic selected.
                R2   ---> Only R2 can be allelic selected.
            
            lines : list
                Information that will be written to the IO stream.
                
        """
        SNP_1 = int(Info[7])
        SNP_2 = int(Info[14])
        
        lines = [Info[1],Info[6],Info[8],Info[13]]
        
        if SNP_1 != 0 and SNP_2 != 0:
            mark = 'Both'
            lines += [mark]
        
        elif SNP_1 != 0 and SNP_2 == 0:
            try:
                candidate = Info[-1]
                SNP_3 = int(Info[21])
                candidate_mark = self._is_Available_Candidate(Info)
                
                if candidate == 'R2' and candidate_mark == True:
                    if SNP_3 != 0:
                        mark = 'Both'
                        lines = [Info[1],Info[6],Info[15],Info[20]]
                        lines += [mark]
                    else:
                        mark = 'R1'
                        lines += [mark]
                else:
                    mark = 'R1'
                    lines += [mark]
            except:
                mark = 'R1'
                lines += [mark]
        
        elif SNP_1 == 0 and SNP_2 != 0:
            try:
                candidate = Info[-1]
                SNP_3 = int(Info[21])
                candidate_mark = self._is_Available_Candidate(Info)

                if candidate== 'R1' and candidate_mark == True:
                    if SNP_3 != 0:
                        mark = 'Both'
                        lines = [Info[15],Info[20],Info[8],Info[13]]
                        lines += [mark]
                    else:
                        mark = 'R2'
                        lines += [mark]
                else:
                    mark = 'R2'
                    lines += [mark]
            except:
                mark = 'R2'
                lines += [mark]
        else:
            #When normal sides cant figure out. Use the Candidate side.
            try:
                candidate = Info[-1]
                candidate_mark = self._is_Available_Candidate(Info)
                SNP_3 = int(Info[21])
                if candidate == 'R1' and SNP_3 != 0 and candidate_mark == True:
                    lines = [Info[15],Info[20],Info[8],Info[13]]
                    mark = 'R1'
                    lines += [mark]
                elif candidate == 'R2' and SNP_3 != 0 and candidate_mark == True:
                    lines = [Info[1],Info[6],Info[15],Info[20]]
                    mark = 'R2'
                    lines += [mark]
                else:
                    mark = 'N'
            except:
                mark = 'N'
        
        if self.save_ID == True:
            return mark, [Info[0]] + lines
        
        else:            
            return mark, lines
                
    
    
    
    
    def Allelic_Filtering(self):
        """
        
        Allelic Tagging.
            
            Pair is Specifical mapped to Paternal Genome:
                if Pair has Paternal SNP:
                    R1 has SNP and R2 has SNP --->   Paternal, Both
                    
                    R1 has SNP and R2 no SNP  --->   Paternal, R1
                    
                    R1 no SNP and R2 has SNP --->    Paternal, R2
                else:
                    Bi-Allelic Pair
            
            Pair is Specifical mapped to Maternal Genome:
                if Pair has Maternal SNP:
                    R1 has SNP and R2 has SNP --->   Maternal, Both
                    
                    R1 has SNP and R2 no SNP  --->   Maternal, R1
                    
                    R1 no SNP and R2 has SNP --->    Maternal, R2
                else:
                    Bi-Allelic Pair
            
            Pair mapped to Both Maternal Genome and Paternal Genome.
                
                =======      =======
                R1                R2
            
            Allelic Tagging is same for two mates.
            
            First,We will check whether mate mapped to the same position on Maternal
            genome and Paternal genome.
                
            Parameters:
            ----------
            pos_M    :    Maternal position
            pos_P    :    Paternal position
            MAX_DIFF_SCORE : the max diff score allowed bewteen two mapping results.
            score_M  :   Mapping punish score on Maternal Genome (AS)
            score_P  :   Mapping punish score on Paternal Genome (AS)
            SNP_M    :   SNP matching sites on Maternal Genome for this mate.
            SNP_P    :   SNP matching sites on Paternal Genome for this mate.
            
            Both Mapping Rules:
            -------------------
            
            if pos_M == pos_P,We first check the SNP matching sites counts. 
                SNP_M > 2 * SNP_P  ---->  Maternal mate.
                2 * SNP_M < SNP_P  ---->  Paternal mate.
            
            if pos_M != pos_P, We first check the difference score. if difference 
            score satisfiy the MAX_DIFF_SCORE.We select the best score position is 
            true position.Then We check the SNP Macthing sites. if nonzero, selected
            the corresponding allelic. 
            
            
            When the rules do not separate the Allelic Reads(Any mates). And the Candidates(C) is
            available. We can Use candidate mate to excavate more potential Allelic
            Information.
            
                ========      ========
                C(M)               C(P)
                
            For Candidates:
                We will replace the traditional mate with the corresponding candidate
                mate. And Repeat the the above process.
                
                
        """
        
        log.log(21,'Sorting for HiC results ...')
        #print 'Sorting for HiC results'
        
        M_bed = self._BigFileSorting(self.Maternal_bed)
        P_bed = self._BigFileSorting(self.Paternal_bed)
        
        log.log(21,'Allelic Filtering ...')
        #print 'Allelic Filtering'
        
        #M_bed = self.Maternal_bed
        #P_bed = self.Paternal_bed
        M_fil = open(M_bed,'r')
        P_fil = open(P_bed,'r')
        #prefix = self.Maternal_bed.split('Maternal')[0]
        #prefix = os.path.split(self.Maternal_bed)[-1].replace('.bed','')
        prefix = os.path.split(self.Maternal_bed)[-1].split('Maternal')[0]+'Valid'
        
        #Out Put
        Bi_Allelic_Out = open(os.path.join(self.Out_Path,prefix+'_Bi_Allelic.bed'),'w')
        M_M_Out = open(os.path.join(self.Out_Path,prefix+'_M_M.bed'),'w')
        P_P_Out = open(os.path.join(self.Out_Path,prefix+'_P_P.bed'),'w')
        M_P_Out = open(os.path.join(self.Out_Path,prefix+'_M_P.bed'),'w')
        P_M_Out = open(os.path.join(self.Out_Path,prefix+'_P_M.bed'),'w') 

        #Statics dic
        
        self.Static_dict = {}
        
        Bi_Allelic = 0
        Both_M = 0
        Both_P = 0
        Single_M = 0
        Single_P = 0
        Regroup = 0   
        Speci_M = 0
        Speci_P = 0
        Speci_M_single = 0
        Speci_M_both = 0
        Speci_P_single = 0
        Speci_P_both = 0
        
        M_line = M_fil.readline()
        P_line = P_fil.readline()
        count = 0 
        
        while True:
            count += 1
            if count % 40000000 == 0:
                log.log(21,'%d pairs completed',count)
                #print '%d pairs completed' % count
            
            self.M_Info = M_line.strip().split()
            self.P_Info = P_line.strip().split()
            
            #All file come to the END. STOP
            if len(self.M_Info) == 0 and len(self.P_Info) == 0:
                break
            
            #Only mapped to Paternal and appear in the END line.
            elif len(self.M_Info) == 0 and len(self.P_Info) != 0:
                Speci_P += 1
                mark, lines = self._Specific_Mapping_line_Process(self.P_Info)
                if mark == 'Both':
                    Both_P += 1
                    Speci_P_both += 1
                    P_P_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                elif mark == 'R1' or mark == 'R2':
                    Single_P += 1
                    Speci_P_single += 1
                    P_P_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                else:
                    Bi_Allelic += 1
                    Bi_Allelic_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                
                P_line = P_fil.readline()

            #Only mapped to Maternal and appear in the END line.
            elif len(self.M_Info) != 0 and len(self.P_Info) == 0:
                Speci_M += 1
                mark, lines = self._Specific_Mapping_line_Process(self.M_Info)
                if mark == 'Both':
                    Both_M += 1
                    Speci_M_both += 1
                    M_M_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                elif mark == 'R1' or mark == 'R2':
                    Single_M += 1
                    Speci_M_single += 1
                    M_M_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                else:
                    Bi_Allelic += 1
                    Bi_Allelic_Out.writelines('\t'.join(self._intToString(lines))+'\n')

                M_line = M_fil.readline()
            
            else:
                M_Name = self.M_Info[0];P_Name = self.P_Info[0]
                
                #Only mapped to Maternal and appears in the middle of file.
                if M_Name < P_Name:
                    Speci_M += 1
                    mark, lines = self._Specific_Mapping_line_Process(self.M_Info)
                    
                    if mark == 'Both':
                        Both_M += 1
                        Speci_M_both += 1
                        M_M_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                    elif mark == 'R1' or mark == 'R2':
                        Single_M += 1
                        Speci_M_single += 1
                        M_M_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                    else:
                        Bi_Allelic += 1
                        Bi_Allelic_Out.writelines('\t'.join(self._intToString(lines))+'\n')

                    M_line = M_fil.readline()
                
                #Only mapped to Paternal and appears in the middle of file.
                elif M_Name > P_Name:
                    Speci_P += 1
                    mark, lines = self._Specific_Mapping_line_Process(self.P_Info)
                    
                    if mark == 'Both':
                        Both_P += 1
                        Speci_P_both += 1
                        P_P_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                    elif mark == 'R1' or mark == 'R2':
                        Single_P += 1
                        Speci_P_single += 1
                        P_P_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                    else:
                        Bi_Allelic += 1
                        Bi_Allelic_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                        
                    P_line = P_fil.readline()
                
                #Pair mapped to the two genome.
                else:
                    mark, lines = self._Both_Mapping_line_Process()
                    if mark == 'NN':
                        Bi_Allelic += 1
                        Bi_Allelic_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                        
                    elif mark == 'NM':
                        Single_M += 1
                        M_M_Out.writelines('\t'.join(self._intToString(lines)+['R2'])+'\n')
                    
                    elif mark == 'MN':
                        Single_M += 1
                        M_M_Out.writelines('\t'.join(self._intToString(lines)+['R1'])+'\n')
                        
                    elif mark == 'MM':
                        Both_M += 1
                        M_M_Out.writelines('\t'.join(self._intToString(lines)+['Both'])+'\n')
                        
                    elif mark == 'NP':
                        Single_P += 1
                        P_P_Out.writelines('\t'.join(self._intToString(lines)+['R2'])+'\n')
                        
                    elif mark == 'PN':
                        Single_P += 1
                        P_P_Out.writelines('\t'.join(self._intToString(lines)+['R1'])+'\n')
                        
                    elif mark == 'PP':
                        Both_P += 1
                        P_P_Out.writelines('\t'.join(self._intToString(lines)+['Both'])+'\n')
                        
                    elif mark == 'MP':
                        Regroup += 1
                        M_P_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                        
                    elif mark == 'PM':
                        Regroup += 1
                        P_M_Out.writelines('\t'.join(self._intToString(lines))+'\n')
                    
                    M_line = M_fil.readline()
                    P_line = P_fil.readline()
        
        log.log(21,"%d pair completed", count-1)
        #print '%d pairs completed' % (count - 1)
        
        #Close the file
        Bi_Allelic_Out.close()
        M_M_Out.close()
        P_P_Out.close()
        M_P_Out.close()
        P_M_Out.close()
        
        #Statics dict
        self.Static_dict['Total_valid_pairs'] = count - 1
        self.Static_dict['Bi_Allelic_pairs'] = Bi_Allelic
        self.Static_dict['Maternal_Allelic_pairs'] = Both_M + Single_M
        self.Static_dict['Paternal_Allelic_pairs'] = Both_P + Single_P
        self.Static_dict['Maternal_both_sides_pairs'] = Both_M
        self.Static_dict['Paternal_both_sides_pairs'] = Both_P
        self.Static_dict['Maternal_single_side_pairs'] = Single_M
        self.Static_dict['Paternal_single_side_pairs'] = Single_P
        self.Static_dict['Speci_Maternal_Mapping_pairs'] = Speci_M
        self.Static_dict['Speci_Paternal_Mapping_pairs'] = Speci_P
        self.Static_dict['Speci_Maternal_both_sides_pairs'] = Speci_M_both
        self.Static_dict['Speci_Paternal_both_sides_pairs'] = Speci_P_both
        self.Static_dict['Speci_Maternal_single_sides_pairs'] = Speci_M_single
        self.Static_dict['Speci_Paternal_single_sides_pairs'] = Speci_P_single
        self.Static_dict['Recombination_pairs'] = Regroup
        self.Static_dict['Allelic_Ratio'] = float(Both_M+Both_P+Single_M+Single_P) / (count-1)
        
        Statics_out = ['Total pairs number is %d' % (count - 1),
                       'Bi-Allelic pairs number is %d' % Bi_Allelic,
                       'Maternal Allelic pairs number is %d' % (Both_M+Single_M),
                       'Paternal Allelic pairs number is %d' % (Both_P+Single_P),
                       'Both sides for Maternal number is %d' % Both_M,
                       'Both sides for Paternal number is %d' % Both_P,
                       'Single sides for Maternal number is %d' % Single_M,
                       'Single sides for Paternal number is %d' % Single_P,
                       'Recombination number is %d' % Regroup,
                       '\n',
                       '==========Specifical Mapping Statics==============',
                       
                       'Specifical Mapping pairs for Maternal number is %d' % Speci_M,
                       '        Both sides for Speci Maternal number is %d' % Speci_M_both,
                       '        Single sides for Speci Maternal number is %d' % Speci_M_single,
                       'Specifical Mapping pairs for Paternal number is %d' % Speci_P,
                       '        Both sides for Speci Paternal number is %d' % Speci_P_both,
                       '        Single sides for Speci Paternal number is %d' % Speci_P_single,
                       '\n',
                       'Allelic Ratio : %.4f' % self.Static_dict['Allelic_Ratio']]
        
        Statics_out = '\n'*2 + '\n'.join(Statics_out) + '\n'
        log.log(21,'Allelic Filtering Results : %s',Statics_out)
        #print 'Allelic Filtering Results : %s' % Statics_out
        
                        
                
                        


# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 15:10:17 2018

@author: han-luo
"""
from __future__ import division
import os, subprocess, multiprocessing, logging, re, sys, time
import xml.etree.ElementTree as ET

log = logging.getLogger(__name__)


def Getchunks(fastq):
    """
        Get chunks files and id from chunk Folder
    """
    chunks =[]
    num = 0
    reg = re.compile(r"(?<=_chunk)\d+")
    for fil in os.listdir(fastq):
        try:
            tmp_num = int(reg.search(fil).group(0))
        except:
            continue
            #log.warning('Unexpected file %s ', fil)
        if tmp_num < num :
            pass
        else:
            num = tmp_num
        
        chunks.append(fil)
    
    Cell_prefix = fil.split('_chunk')[0]
    
    return chunks, num + 1, Cell_prefix
    

#-------------------------Global Mapping Functions----------------------------------
    
def WS_sub_mapping(bowtieIndex, bowtiePath, threads, fq, OutPath, task_count):
    
    """
        sub-mapping  for WS station
        
    Parameters
    ----------
    bowtieIndex : str
        bowtieIndex if you already build the genome Index.
    
    bowtiePath : str
        bowtie2 path.
    
    threads : int
        threads
    
    fq : str
        Sequencing data for any chunk of one mate.
    
    OutPath : str
        OutFolder for mapping results.
    
    task_count : int
        id of sub-mapping task
    """
    
    fq_prefix = os.path.split(fq)[-1].split('.')[0]
    genome_prefix = os.path.split(bowtieIndex)[-1]
    out_sam_name = fq_prefix + '_' + genome_prefix + '.sam'
    out_sam_file = os.path.join(OutPath,out_sam_name)
    out_bam_file = out_sam_file.rstrip('.sam') + '.bam'
    prefix = out_sam_file.rstrip('.sam')
    
    mapcmd = [bowtiePath, '-x', bowtieIndex, '-p', str(threads), '-U', fq, '|', 
              'samtools', 'view', '-b','-S', '|', 
              'samtools', 'sort', '-n','-T', prefix, '-o', out_bam_file]
    #log.log(21,'%s',' '.join(mapcmd))
    map_res = subprocess.Popen(' '.join(mapcmd),shell=True,stderr=subprocess.PIPE,bufsize=-1)
    
    map_res.communicate()
    
#    out_bam_file = out_sam_file.rstrip('.sam') + '.bam'
#    prefix = out_sam_file.rstrip('.sam')
#    sam_precess = ['samtools', 'view', '-b','-S',out_sam_file, '|', 'samtools', 'sort', '-n','-T', prefix, '-o', out_bam_file]
#    #log.log(21,'%s',' '.join(sam_precess))
#    samtools_res = subprocess.Popen(' '.join(sam_precess),shell = True,stdout=subprocess.PIPE,
#                                    stderr=subprocess.PIPE,bufsize=-1)
#    samtools_res.communicate()
    
    log.log(21,'%s mapping to %s done! count_id : %d ',fq_prefix,genome_prefix,task_count)
    


def WS_mapping(fastq, threads, bowtiePath, OutPath, bowtieIndex = None):
    """
        WS(work station) Mapping mode.
        submit the mapping task in automatic way of system.
    
    Parameters
    ----------
    
    fastq : str
        chunk fq.gz Folder created by update.
        
    threads : str
        Total number of threads for four sub-mapping tasks.

    bowtiePath : str
        bowtie2 Path
        
    bowtieIndex : list or tuple
        List or tuple for Index of Maternal & Paternal Genome
    
    OutPath : str
        sam OutPath.
               
    """
    if not (isinstance(bowtieIndex,list) or isinstance(bowtieIndex,tuple)):
        log.error('bowtieIndex arrangement error ...')
        log.error('Exit ...')
        sys.exit(1)
    elif len(bowtieIndex) == 2:
        log.log(21,'two Index were found, arrange for M/Paternal genome ...')
        Maternal_Index = bowtieIndex[0]
        Paternal_Index = bowtieIndex[1]
        log.log(21,'Maternal Genome Index is %s ',bowtieIndex[0])
        log.log(21,'Paternal Genome Index is %s ',bowtieIndex[1])
    
        sub_threads = threads // 4
        chunks, num, Cell = Getchunks(fastq)
        log.log(21,'%d chunk paris were found under %s',num,fastq)
        log.log(21,'total number of sub-mapping tasks is %d ', num*4)    
    
        mapping_Pool = multiprocessing.Pool(4)
        count = 0
        for fil in chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            mapping_Pool.apply_async(WS_sub_mapping,args=(Maternal_Index,
                                                          bowtiePath,
                                                          sub_threads,
                                                          fq,
                                                          OutPath,
                                                          count))
            count += 1
            mapping_Pool.apply_async(WS_sub_mapping,args=(Paternal_Index,
                                                          bowtiePath,
                                                          sub_threads,
                                                          fq,
                                                          OutPath,
                                                          count))
            
        time.sleep(2)
        mapping_Pool.close()
        mapping_Pool.join()        
    elif len(bowtieIndex) == 1:
        log.log(21,'Only one Index was found, Non-Allelic HiC mapping ...')
        genome_Index = bowtieIndex[0]
        log.log(21,'Genome Index is %s ', genome_Index)
        
        sub_threads = threads // 4
        chunks, num, Cell = Getchunks(fastq)
        log.log(21,'%d chunk paris were found under %s',num,fastq)
        log.log(21,'total number of sub-mapping tasks is %d ', num*2)
        
        mapping_Pool = multiprocessing.Pool(4)
        count = 0
        for fil in chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            mapping_Pool.apply_async(WS_sub_mapping, args = (genome_Index,
                                                             bowtiePath,
                                                             sub_threads,
                                                             fq,
                                                             OutPath,
                                                             count))
        time.sleep(2)
        mapping_Pool.close()
        mapping_Pool.join()
        
    else:
        log.error('bowtieIndex Number not correct, only one Index or two Index supported. ...')
        log.error('Exit')
        sys.exit(1)
    
    log.log(21,'Multiprocess mapping Done!'
                'Try to continue with filtering task,haphic filtering -h for help information')
    
    
 
def PBS_sub_mapping(fq, bowtieIndex, bowtiePath, threads, OutPath, logPath, mem,task_count):
    """
        PBS scripts for sub mapping.
    
    Parameters
    ----------
    
    fq : str
        Sequencing data for any chunk of one mate.
    
    bowtieIndex : str
        Genome Index Path if you already build the genome Index.
    
    bowtiePath : str
        bowtie2 Path.
        
    threads : int
        threads.
    
    OutPath : str
        OutFolder for mapping results.
        
    logPath : str
        OutFolder for mapping log.
    
    mem : int
        memory  resource for each task        
                    
    task_count : int
        id of sub-mapping task
        
    """
    cell = os.path.split(fq)[-1].split('_chunk')[0]
    fq_prefix = os.path.split(fq)[-1].split('.')[0]
    genome_prefix = os.path.split(bowtieIndex)[-1]
    out_bam_name = fq_prefix + '_' + genome_prefix + '.bam'
    prefix = out_bam_name.rstrip('.bam')
    out_bam_file = os.path.join(OutPath,out_bam_name)
    
    PBS_scripts = ('echo "bowtie2 -x %s -U %s -p %s  | samtools view -b -S | '
                    ' samtools sort -n -T %s -@ %s -o %s" '
                    '| qsub -N %s '
                    '-l nodes=1:ppn=%s -l mem=%sgb -d ./ '
                    '-e %s -o %s'
                    %(bowtieIndex,fq,threads,prefix,threads,out_bam_file,
                      cell ,threads,mem,logPath,logPath))
    
    subprocess.Popen(PBS_scripts,shell=True,stdout=subprocess.PIPE)
    
    log.log(21,'%s To %s has been submitted. count_id : %d ',fq_prefix, genome_prefix, task_count)
    
#def PBS_Pooler(keyword):
#    """
#        Check the tasks state in Pool.
#    
#    Parameters
#    ----------
#    keyword : str
#        key word of task name. 
#    
#    return the number of task which has key word.      
#    """
#    
#    caller = subprocess.check_output(['qstat'])
#    
#    state = caller.split('\n')[2:-1]
#    
#    num = 0
#    for i in state:
#        i = i.split()
#        if keyword in i[1]:
#            num += 1
#        else:
#            pass
#    
#    return num

def PBS_Pooler(keyword):
    """
        Check the tasks state in Pool.
    
    Parameters
    ----------
    keyword : str
        key word of task name.
    
    return the number of task which has keyword.
    """
    
    jobs_XML = os.popen('qstat -xl') # check the jobs full information as XML format
    job_num = 0
    try:
        tree = ET.parse(jobs_XML)
    except:
        return 0
    root = tree.getroot()
    
    for sub_job in root.getchildren():
        tmp_name = sub_job.find('Job_Name').text
        if keyword in tmp_name:
            job_num += 1
        else:
            continue
    return job_num



def Pool_controller(keyword, num_task):
    """
        controller of Whether submit the new task.
    """
    while True:
        if PBS_Pooler(keyword) < num_task:
            return True
        else:
            time.sleep(5)

def check_Non_Allelic_out(path,chunk_num):
    """
        check out whether the mapping output are completed.
        we will check whether there are 2 mates mapping sam/bam for a chunk id.
        and whether the size of mapping results are zero.
        
    Parameters
    ----------
    
    path : str
        mapping out path
    
    chunk_num : int
        The number of chunks.    
        
    """
    files = os.listdir(path)
    error_list = []
    for i in range(chunk_num):
        tmp = []
        for fil in files:
            if 'chunk'+str(i)+'_' in fil:
                tmp.append(fil)
        if len(tmp) != 2:
            R1 = 'chunk'+str(i)+'_1'
            R2 = 'chunk'+str(i)+'_2'
            R1_tmp = False
            R2_tmp = False
            for fil in tmp:
                if R1 in fil:
                    R1_tmp = True
                if R2 in fil:
                    R2_tmp = True
            if not R1_tmp:
                error_list.append((i,'1'))
                log.warning('chunk%s_1 mapping results is lost...',i)
            if not R2_tmp:
                error_list.append((i,'2'))
                log.warning('chunk%s_2 mapping results is lost...',i)
    
    for fil in files:
        if os.path.getsize(os.path.join(path,fil)) < 100:
            tmp = fil.split('chunk')[1].replace('.bam','').split('_')[:-1]
            error_list.append(tuple(tmp))
            log.warning('mapping results %s is empty',fil)

    return error_list
    
    
def check_Allelic_out(path,chunk_num):
    """
        check out whether the mapping output are completed.
        We will check whether there are 4 Allelic mapping sam/bam for a chunk id.
        and whether the size of mapping results are zero.
    
    Parameters
    ----------
    
    path : str
        mapping out path
    
    chunk_num : int
        The number of chunks.
        
    """
    files = os.listdir(path)
    error_list = []
    for i in range(chunk_num):
        tmp = []
        for fil in files:
            if 'chunk'+str(i)+'_' in fil:
                tmp.append(fil)
        if len(tmp) != 4:
            M_1 = 'chunk'+str(i)+'_1_Maternal'
            M_2 = 'chunk'+str(i)+'_2_Maternal'
            P_1 = 'chunk'+str(i)+'_1_Paternal'
            P_2 = 'chunk'+str(i)+'_2_Paternal'
            M_1_tmp = False
            M_2_tmp = False
            P_1_tmp = False
            P_2_tmp = False
            for fil in tmp:
                if M_1 in fil:
                    M_1_tmp = True
                if M_2 in fil:
                    M_2_tmp = True
                if P_1 in fil:
                    P_1_tmp = True
                if P_2 in fil:
                    P_2_tmp = True
            if not M_1_tmp:
                error_list.append((i,'1','Maternal'))
                log.warning('chunk%s_1 mapping To Maternal is lost...',i)
            if not M_2_tmp:
                error_list.append((i,'2','Maternal'))
                log.warning('chunk%s_2 mapping To Maternal is lost...',i)
            if not P_1_tmp:
                error_list.append((i,'1','Paternal'))
                log.warning('chunk%s_1 mapping To Paternal is lost...',i)
            if not P_2_tmp:
                error_list.append((i,'2','Paternal'))
                log.warning('chunk%s_2 mapping To Paternal is lost...',i)
            
    
    for fil in files:
        if os.path.getsize(os.path.join(path,fil)) < 100:
            tmp = fil.split('chunk')[1].rstrip('.bam').split('_')
            error_list.append(tuple(tmp))
            log.error('mapping results %s is empty',fil)
      
    return error_list
    

def PBS_controller(fastq, num_task, threads, OutPath, logPath, bowtiePath, mem, bowtieIndex = None):   
    """
        PBS tasks controller.
    Parameters
    ----------
    fastq : str
        chunk fq.gz Folder created by update.
        
    num_task : int
        Total sub-tasks number in task Pools.
    
    threads : int
        number of threads for a sub mapping task.
    
    OutPath : str
        OutFolder for mapping results.
    
    logPath : str
        OutFolder for mapping log.
    
    bowtiePath : str
        bowtie2 Path

    mem : int
        memory for each sub mapping tasks.        
        
    bowtieIndex : list or tuple
        List or tuple for Index of Maternal & Paternal Genome.
    
    """
    if not (isinstance(bowtieIndex,list) or isinstance(bowtieIndex,tuple)):
        log.error('Allelic bowtieIndex arrangement error ...')
        log.error('Exit ...')
        sys.exit(1)
    elif len(bowtieIndex) == 2:
        log.log(21,'two Index were found, arrange for M/Paternal genome ...')
        Maternal_Index = bowtieIndex[0]
        Paternal_Index = bowtieIndex[1]
        log.log(21,'Maternal Genome Index is %s ',bowtieIndex[0])
        log.log(21,'Paternal Genome Index is %s ',bowtieIndex[1])    
    
        chunks, chunk_num, Cell = Getchunks(fastq)
        log.log(21,'%d chunk paris were found under %s',chunk_num,fastq)
        log.log(21,'total number of sub-mapping tasks is %d ', chunk_num * 4)    
    
        count = 0
        for fil in chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            if Pool_controller(Cell, num_task):
                PBS_sub_mapping(fq=fq,
                                bowtieIndex = Maternal_Index,
                                bowtiePath= bowtiePath,
                                threads = threads,
                                OutPath = OutPath,
                                logPath=logPath,
                                mem = mem,
                                task_count= count)
            time.sleep(2)
            count += 1        
            if Pool_controller(Cell, num_task):
                PBS_sub_mapping(fq=fq,
                                bowtieIndex = Paternal_Index,
                                bowtiePath= bowtiePath,
                                threads = threads,
                                OutPath = OutPath,
                                logPath=logPath,
                                mem = mem,
                                task_count= count)
            time.sleep(2)
    
        while True:
            if PBS_Pooler(Cell) == 0:
                logging.log(21,'all chunks have been mapped To the Genome.')
                break
            else:
                time.sleep(2)
    
        # re-building mapping task for error mapping results.
        while True:
            error_list = check_Allelic_out(OutPath,chunk_num)
            if len(error_list) == 0:
                log.log(21,'all chunks mapping results are completed.')
                break
            else:
                log.log(21,'rebuilding the mapping for escaped mapping tasks')
                count = 0
                for i in error_list:
                    mask = 'chunk'+str(i[0])+'_'+str(i[1])
                    fq = [fil for fil in chunks if mask in fil][0]
                    fq = os.path.join(fastq,fq)
                    count += 1
                    if i[2] == 'Maternal':
                        Index = Maternal_Index
                    else:
                        Index = Paternal_Index
            
                    if Pool_controller(Cell, num_task):
                        PBS_sub_mapping(fq=fq,
                                        bowtieIndex = Index,
                                        bowtiePath= bowtiePath,
                                        threads = threads,
                                        OutPath = OutPath,
                                        logPath=logPath,
                                        mem = mem,
                                        task_count= count)
                    time.sleep(2)
        
            log.log(21,'all rebuilding tasks have submitted.')
    
            while True:
                if PBS_Pooler(Cell) == 0:
                    logging.log(21,'all rebuilding tasks have been mapped To the Genome.')
                    break
                else:
                    time.sleep(5)
    elif len(bowtieIndex) == 1:
        log.log(21,'Only one Index was found, Non-Allelic HiC mapping ...')
        genome_Index = bowtieIndex[0]
        log.log(21,'Genome Index is %s ', genome_Index)
        chunks, chunk_num, Cell = Getchunks(fastq)
        log.log(21,'%d chunk paris were found under %s',chunk_num,fastq)
        log.log(21,'total number of sub-mapping tasks is %d ', chunk_num*2)
        
        count = 0
        for fil in chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            if Pool_controller(Cell, num_task):
                PBS_sub_mapping(fq = fq,
                                bowtieIndex=genome_Index,
                                bowtiePath=bowtiePath,
                                threads=threads,
                                OutPath=OutPath,
                                logPath=logPath,
                                mem=mem,
                                task_count=count)
            time.sleep(2)
        while True:
            if PBS_Pooler(Cell) == 0:
                logging.log(21,'all chunks have been mapped To the Genome.')
                break
            else:
                time.sleep(5)
        
        # re-building mapping task for error mapping results
        while True:
            error_list = check_Non_Allelic_out(OutPath,chunk_num)
            if len(error_list) == 0:
                log.log(21,'all chunks mapping results are completed.')
                break
            else:
                log.log(21,'rebuilding the mapping for escaped mapping tasks')
                count = 0
                for i in error_list:
                    mask = 'chunk'+str(i[0])+'_'+str(i[1])
                    fq = [fil for fil in chunks if mask in fil][0]
                    fq = os.path.join(fastq,fq)
                    count += 1
                    if Pool_controller(Cell, num_task):
                        PBS_sub_mapping(fq = fq,
                                        bowtieIndex = genome_Index,
                                        bowtiePath = bowtiePath,
                                        threads = threads,
                                        OutPath = OutPath,
                                        logPath = logPath,
                                        mem = mem,
                                        task_count = count)
                    time.sleep(2)
            
            log.log(21,'all rebuilding tasks have submitted.')
            
            while True:
                if PBS_Pooler(Cell) == 0:
                    logging.log(21,'all rebuilding tasks have been mapped To the Genome.')
                    break
                else:
                    time.sleep(5)
                                        
    else:
        log.error('bowtieIndex Number not correct, only one Index or two Index supported. ...')
        log.error('Exit')
        sys.exit(1)

    
#----------------------------------Rescue Mapping ----------------------------------

def Rescue_sub_WS_mapping(bowtieIndex, bowtiePath, threads, fq, OutPath):
    """
        sub-rescue-mapping for WS station.
    
    Parameters
    ----------
    bowtieIndex : str
        bowtieIndex if you already build the genome Index.
    
    bowtiePath : str
        bowtie2 Path.
    
    threads : int
        threads
    
    fq : str
        Sequencing data for any chunk of one mate.
    
    OutPath : str
        OutFolder for mapping results
        
    """
    fq_prefix = os.path.join(OutPath,os.path.split(fq)[-1].split('.')[0])
    out_bam_file = os.path.join(OutPath,fq_prefix+'.bam')
    
    mapcmd = [bowtiePath, '-x', bowtieIndex, '-p', str(threads), '-U', fq, '|', 
              'samtools', 'view', '-b','-S', '|', 
              'samtools', 'sort', '-n','-T', fq_prefix, '-o', out_bam_file] 
              
    map_res = subprocess.Popen(' '.join(mapcmd), shell=True,stderr=subprocess.PIPE,bufsize=-1)
    
    #print ' '.join(mapcmd)
    map_res.communicate()
    

 
def Rescue_WS_mapping(fastq,threads,bowtiePath,OutPath,bowtieIndex = None):
    """
        Resuce mapping for WS station.
        
    """
    if not (isinstance(bowtieIndex,list) or isinstance(bowtieIndex,tuple)):
        log.error('BowtieIndex arrangement error ...')
        log.error('Exit ...')
        sys.exit(1)
    elif len(bowtieIndex) == 2:
        log.log(21,'two Index were found, arrange for M/Paternal genome ...')
        log.log(21,'Allelic Rescue Mapping start ...')
        Maternal_Index = bowtieIndex[0]
        Paternal_Index = bowtieIndex[1]
        log.log(21,'Maternal Genome Index is %s', bowtieIndex[0])
        log.log(21,'Paternal Genome Index is %s', bowtieIndex[1])
        
        sub_threads = threads // 4
        Maternal_chunks = [i for i in os.listdir(fastq) if 'Maternal' in i]
        Paternal_chunks = [i for i in os.listdir(fastq) if 'Paternal' in i]
        
        log.log(21,'Rescue Mapping ...')
        mapping_Pool = multiprocessing.Pool(4)
        for fil in Maternal_chunks:
            fq = os.path.join(fastq,fil)
            
            mapping_Pool.apply_async(Rescue_sub_WS_mapping,args=(Maternal_Index,
                                                                 bowtiePath,
                                                                 sub_threads,
                                                                 fq,
                                                                 OutPath))
        time.sleep(1)
        for fil in Paternal_chunks:
            fq = os.path.join(fastq,fil)
            mapping_Pool.apply_async(Rescue_sub_WS_mapping,args=(Paternal_Index,
                                                                 bowtiePath,
                                                                 sub_threads,
                                                                 fq,
                                                                 OutPath))
        time.sleep(1)
        mapping_Pool.close()
        mapping_Pool.join()
        log.log(21,'Rescue Mapping Done!')
        
    elif len(bowtieIndex) == 1:
        log.log(21,'Only one Index was found, Non-Allelic Rescue Mapping start...')
        genome_Index = bowtieIndex[0]
        log.log(21,'Genome Index is %s ', genome_Index)
        
        sub_threads = threads // 4
        fastq_chunks = [i for i in os.listdir(fastq) if 'chunk' in i]

        log.log(21,'Rescue Mapping')
        mapping_Pool = multiprocessing.Pool(4)
        for fil in fastq_chunks:
            fq = os.path.join(fastq,fil)
            mapping_Pool.apply_async(Rescue_sub_WS_mapping,args=(genome_Index,
                                                                 bowtiePath,
                                                                 sub_threads,
                                                                 fq,
                                                                 OutPath))
        time.sleep(1)
        mapping_Pool.close()
        mapping_Pool.join()
        log.log(21,'Rescue Mapping Done!')
    else:
        log.error('bowtieIndex Number not correct, only one Index or two Index supported. ...')
        log.error('Exit ...')
        sys.exit(1)


def Rescue_check_out(fq,OutPath):
    """
        Check Out Whether the ReMapping results.  
    """
    errorlist = []
    Info_fq = os.listdir(fq)
    Info_bam = os.listdir(OutPath)
    for i in Info_fq:
        tmp = i.split('.')[0]+'.bam'
        if tmp in Info_bam:
            if os.path.getsize(os.path.join(OutPath,tmp)) < 100:
                error = i.split('chunk')[-1].split('_')[:3]
                errorlist.append(error)
                log.warning('chunk%s_%s to %s ReMapping result is empty ...',error[0],error[1],error[2])
                #print 'chunk%s_%s to %s ReMapping result is empty ...' % (error[0],error[1],error[2])
            else:
                pass
        else:
            error = i.split('chunk')[-1].split('_')[:3]
            errorlist.append(error)
            log.warning('chunk%s_%s to %s ReMapping result is lost ...',error[0],error[1],error[2])
            #print 'chunk%s_%s to %s ReMapping result is lost ...'% (error[0],error[1],error[2])
    return errorlist





def Rescue_sub_PBS_mapping(fq,bowtieIndex,bowtiePath,threads,OutPath,logPath,mem,task_count):
    """
        PBS scripts for sub Rescue mapping.
        
    Parameters
    ----------
    fq : str
        Sequencing data for any chunk of one mate.
    
    bowtieIndex : str
        Genome Index Path if you already build the genome Index.
    
    bowtiePath : str
        bowtie2 Path.
        
    threads : int
        threads.
    
    OutPath : str
        OutFolder for mapping results.
        
    logPath : str
        OutFolder for mapping log.
    
    mem : int
        memory  resource for each task        
                    
    task_count : int
        id of sub-mapping task
        
    """
    cell = os.path.split(fq)[-1].split('_chunk')[0]
    fq_prefix = os.path.split(fq)[-1].split('.')[0]
    out_bam_file = os.path.join(OutPath,fq_prefix+'.bam')
    
    PBS_scripts = ('echo "bowtie2 -x %s -U %s -p %s  | samtools view -b -S | '
                    ' samtools sort -n -T %s -@ %s -o %s" '
                    '| qsub -N %s '
                    '-l nodes=1:ppn=%s -l mem=%sgb -d ./ '
                    '-e %s -o %s'
                    %(bowtieIndex,fq,threads,fq_prefix,threads,out_bam_file,
                      cell ,threads,mem,logPath,logPath))
    
    subprocess.Popen(PBS_scripts,shell=True,stdout=subprocess.PIPE)
    log.log(21,'%s  is remapping . count_id : %d ',fq_prefix, task_count)
    

def Rescue_PBS_controller(fastq,num_task, threads, OutPath, logPath, bowtiePath, mem, bowtieIndex = None):
    """
       Rescue PBS tasks controller.
    Parameters
    ----------
    fastq : str
        chunk fq.gz Folder created by update.
        
    num_task : int
        Total sub-tasks number in task Pools.
    
    threads : int
        number of threads for a sub mapping task.
    
    OutPath : str
        OutFolder for mapping results.
    
    logPath : str
        OutFolder for mapping log.
    
    bowtiePath : str
        bowtie2 Path

    mem : int
        memory for each sub mapping tasks.        
        
    bowtieIndex : list or tuple
        List or tuple for Index of  Genome.    
        
    """
    
    if not (isinstance(bowtieIndex,list) or isinstance(bowtieIndex,tuple)):
        log.error('bowtieIndex arrangement error')
        log.error('Exit ...')
        sys.exit(1)
    elif len(bowtieIndex) == 2:
        log.log(21,'two Index were found, arrange for M/Paternal genome ...')
        Maternal_Index = bowtieIndex[0]
        Paternal_Index = bowtieIndex[1]
        log.log(21,'Maternal Genome Index is %s', bowtieIndex[0])
        log.log(21,'Paternal Genome Index is %s', bowtieIndex[1])
        
        count = 0
        Maternal_chunks = [i for i in os.listdir(fastq) if 'Maternal' in i]
        Paternal_chunks = [i for i in os.listdir(fastq) if 'Paternal' in i]
        Cell = Maternal_chunks[0].split('_chunk')[0]
        for fil in Maternal_chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            if Pool_controller(Cell, num_task):
                Rescue_sub_PBS_mapping(fq=fq,
                                       bowtieIndex=Maternal_Index,
                                       bowtiePath=bowtiePath,
                                       threads=threads,
                                       OutPath=OutPath,
                                       logPath=logPath,
                                       mem=mem,
                                       task_count=count)
            time.sleep(1)
        
        for fil in Paternal_chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            if Pool_controller(Cell,num_task):
                Rescue_sub_PBS_mapping(fq=fq,
                                       bowtieIndex=Paternal_Index,
                                       bowtiePath=bowtiePath,
                                       threads=threads,
                                       OutPath=OutPath,
                                       logPath=logPath,
                                       mem=mem,
                                       task_count=count)
            time.sleep(1)
        
        while True:
            if PBS_Pooler(Cell) == 0:
                log.log(21,'Rescue Mapping have Done,Check the reults ...')
                break
            else:
                time.sleep(5)
        
        #re-building ReMapping task for error ReMapping results
        while True:
            error_list = Rescue_check_out(fastq,OutPath)
            if len(error_list) == 0:
                log.log(21,'all chunks mapping results are completed.')
                break
            else:
                log.log(21,'rebuilding the mapping for escaped mapping tasks')
                count = 0
                for i in error_list:
                    mask = 'chunk'+str(i[0])+'_'+str(i[1])
                    if i[2] == 'Maternal':
                        fq = [fil for fil in Maternal_chunks if mask in fil][0]
                        fq = os.path.join(fastq,fq)                        
                        bowtieIndex = Maternal_Index
                    else:
                        fq = [fil for fil in Paternal_chunks if mask in fil][0]
                        fq = os.path.join(fastq,fq)
                        bowtieIndex = Paternal_Index
                    count += 1
                    if Pool_controller(Cell, num_task):
                        Rescue_sub_PBS_mapping(fq = fq,
                                               bowtieIndex = bowtieIndex,
                                               bowtiePath = bowtiePath,
                                               threads = threads,
                                               OutPath = OutPath,
                                               logPath = logPath,
                                               mem = mem,
                                               task_count = count)
                    time.sleep(2)
            log.log(21,'all rebuilding tasks have submitted')
            
            while True:
                if PBS_Pooler(Cell) == 0:
                    logging.log(21,'all rebuilding tasks have been mapped To the Genome.')
                    break
                else:
                    time.sleep(5)
    
    elif len(bowtieIndex) == 1:
        log.log(21,'Only one Index was found, Non-Allelic Rescue Mapping start...')
        genome_Index = bowtieIndex[0]
        log.log(21,'Genome Index is %s', genome_Index)
        fastq_chunks = [i for i in os.listdir(fastq) if 'chunk' in i]
        Cell = fastq_chunks[0].split('_chunk')[0]
        
        count = 0
        for fil in fastq_chunks:
            fq = os.path.join(fastq,fil)
            count += 1
            if Pool_controller(Cell,num_task):
                Rescue_sub_PBS_mapping(fq=fq,
                                       bowtieIndex=genome_Index,
                                       bowtiePath=bowtiePath,
                                       threads=threads,
                                       OutPath=OutPath,
                                       logPath=logPath,
                                       mem=mem,
                                       task_count=count)
            time.sleep(1)
        
        while True:
            if PBS_Pooler(Cell) == 0:
                log.log(21,'Rescue Mapping have Done,Check the results')
                break
            else:
                time.sleep(5)
        #rebuilding the ReMapping task for error task
        while True:
            error_list = Rescue_check_out(fastq,OutPath)
            if len(error_list) == 0:
                log.log(21,'all chunks mapping results are completed.')
                break
            else:
                log.log(21,'rebuilding the mapping for escaped mapping tasks')
                count = 0
                for i in error_list:
                    mask = 'chunk'+str(i[0])+'_'+str(i[1])
                    fq = [fil for fil in fastq_chunks if mask in fil][0]
                    fq = os.path.join(fastq,fq)
                    bowtieIndex = genome_Index
                    count += 1
                    if Pool_controller(Cell, num_task):
                        Rescue_sub_PBS_mapping(fq = fq,
                                               bowtieIndex = bowtieIndex,
                                               bowtiePath = bowtiePath,
                                               threads = threads,
                                               OutPath = OutPath,
                                               logPath = logPath,
                                               mem = mem,
                                               task_count = count)
                    time.sleep(2)
            log.log(21,'all rebuilding tasks have submitted')
            
            while True:
                if PBS_Pooler(Cell) == 0:
                    logging.log(21,'all rebuilding tasks have been mapped To the Genome.')
                    break
                else:
                    time.sleep(5)




# -*- coding: utf-8 -*-
"""
Created on Wed Feb 28 09:26:36 2018

@author: han-luo
"""
import numpy as np
import os, logging, re, subprocess, bisect, cPickle
import Bio
import Bio.Restriction
Bio.Restriction         #shut up the editor warning

from HiCHap.fastqPlus import Enzyme_Handle

log = logging.getLogger(__name__)



def buildIndex(genome, OutPath, threads):
    """
        Build Genome Index.
    """
    genome_fil = os.path.split(genome)[-1]
    bowtie_Index = os.path.join(OutPath,genome_fil.rstrip('.fa'))
    log.log(21,'Building Index %s...', bowtie_Index)
    
    buildcmd = ['bowtie2-build','--threads',str(threads), genome, bowtie_Index]
    log.log(21,'command : %s',' '.join(buildcmd))
    subprocess.call(buildcmd)
    log.log(21,'BowtieIndex %s, Done', bowtie_Index)
    

    
def enzymeFind(genome, enzymeName, OutPath):
    """
        Finds restriction sites.
        Create a enzyme fragment file.Format and rule like :
        eg : 
            seq : 1(chr)  CTAGATCAATTCGATCTTAC    enzyme : GATC
            the file will be :
            1 1 4    ->   [1,4)
            1 4 13   ->   [4,13) 
            1 13 20  ->   [13,20)
        
    Parameters
    ----------
    genome : str
        Original genome file
    
    enzymeName : str
        enzyme Site
        
    """
    
    # loading Genome        
    dict_genome = {}
    file_genome = open(genome,'r')
    genome_name = genome.split('/')[-1].rstrip('.fa')
    for line in file_genome:
        line = line.strip('\n')
        lists = list(line)
        if len(lists) > 1 and lists[0] == '>':
            chrs = (line.split('>')[1]).split()[0].lstrip('chr')
            dict_genome[chrs] = []
        else:
            dict_genome[chrs].extend(lists)
    # Finding enzyme site

    enzyme_site, cutsite = Enzyme_Handle(enzymeName)
    
    enzyme_file = os.path.join(OutPath,enzymeName+'_'+genome_name+'_fragments.txt')
    f = open(enzyme_file,'w')
    for k in sorted(dict_genome.keys()):
        seq_str = ''.join(dict_genome[k]).upper()
        pos_list = [int(m.start() + 1 + cutsite[0]) for m in re.finditer(enzyme_site, seq_str)]
        pos_list = [1] + pos_list + [len(seq_str)]
        for i in range(len(pos_list)-1):
            line = [k, str(pos_list[i]), str(pos_list[i+1])]
            f.writelines('\t'.join(line)+'\n')
    f.close()






def SNPs_integration(SNP_file,outPath):
    """
        Integrate The SNP information.
    """
    
    log.log(21,'Loading SNPs And  creat temp file under %s',outPath)
    Whole_SNPs = {}
    SNP_f = open(SNP_file,'r')
    for line in SNP_f:
        line = line.split()
        if line[0] not in Whole_SNPs.keys():
            Whole_SNPs[line[0]] = {}
            Whole_SNPs[line[0]]['pos'] = [int(line[1])]
            Whole_SNPs[line[0]]['ref'] = [line[2]]
            Whole_SNPs[line[0]]['m_alt'] = [line[3]]
            Whole_SNPs[line[0]]['p_alt'] = [line[4]]
        else:
            mid_pos = bisect.bisect_left(Whole_SNPs[line[0]]['pos'],int(line[1]))
            Whole_SNPs[line[0]]['pos'].insert(mid_pos,int(line[1]))
            Whole_SNPs[line[0]]['ref'].insert(mid_pos,line[2])
            Whole_SNPs[line[0]]['m_alt'].insert(mid_pos,line[3])
            Whole_SNPs[line[0]]['p_alt'].insert(mid_pos,line[4])
    
    for c in Whole_SNPs.keys():
        Whole_SNPs[c]['pos'] = np.array(Whole_SNPs[c]['pos'])
        Whole_SNPs[c]['ref'] = np.array(Whole_SNPs[c]['ref'])
        Whole_SNPs[c]['p_alt'] = np.array(Whole_SNPs[c]['p_alt'])
        Whole_SNPs[c]['m_alt'] = np.array(Whole_SNPs[c]['m_alt'])

    f = open(os.path.join(outPath,'Snps.pickle'),'wb')
    cPickle.dump(Whole_SNPs,f,2)
    f.close()
    #np.savez_compressed(os.path.join(outPath,'Snps.pickle'),**Whole_SNPs)
    log.log(21,'Done!,temp Snps file %s',os.path.join(outPath,'Snps.pickle'))




def OutputGenome(dict_genome,outfile):
    """
    """
    file_out = open(outfile,'w')
    for k in sorted(dict_genome.keys()):
        out_list = dict_genome[k]
        lens = len(out_list)
        n = int(lens) / 60 + 1
        file_out.write('>'+'chr'+k+' dna:chromosome chromosome:HapHiC:1:1:'\
                        +str(lens)+':1 REF'+'\n')
        for i in range(n):
            strs = ''.join(out_list[(i*60) : ((i+1)*60)])
            file_out.write(strs+'\n')


def buildRawGenome(Input, enzyme, OutPath, threads):
    """
        Build the raw genome Index, A simple way to non-Allelic HiC process.    
    
    """
    log.log(21,'Generate the genome size file')
    dict_genome = {}
    file_genome = open(Input,'r')
    for line in file_genome:
        line = line.strip('\n')
        lists = list(line)
        if len(lists) > 1 and lists[0] == '>':
            chrs = (line.split('>')[1]).split()[0].lstrip('chr')
            dict_genome[chrs] = []
        else:
            dict_genome[chrs].extend(lists)
    
    #Generate the genomeSzie file
    with open(os.path.join(OutPath,'genomeSize'),'w') as o:
        for key in sorted(dict_genome.keys()):
            o.writelines('\t'.join([key,str(len(dict_genome[key]))])+'\n')
    
    del dict_genome
    log.log(21,'split the genome %s by enzyme.',Input)
    enzymeFind(Input,enzyme,OutPath)
    
    buildIndex(Input, OutPath, threads)
    log.log(21,'genome %s done!',Input)



def rebuildGenome(genomePath, snpPath, enzyme, OutPath, threads):
    """
    Rebuild the genome by replacing the snp sites. And build BowtieIndex.
    
    Parameters
    ----------
    genomePath : str
        Original genome file.
        
    snpPath : str
        SNP file.
        
    enzyme : str
        enzyme site.
        
    OutPath : str
        Rebuild Genome Folder
    
    """
    #Load the snp
    log.log(21,'Loading SNP from temp Snp file...')
    
    Snps = cPickle.load(open(snpPath,'rb'))
        
    #Load the Genome
    log.log(21,'Loading Genome ...')
    dict_genome = {}
    file_genome = open(genomePath,'r')
    for line in file_genome:
        line = line.strip('\n')
        lists = list(line)
        if len(lists) > 1 and lists[0] == '>':
            chrs = (line.split('>')[1]).split()[0].lstrip('chr')
            dict_genome[chrs] = []
        else:
            dict_genome[chrs].extend(lists)
    
    #Generate the genomeSzie file
    log.log(21,'Generate the genome size file')
    with open(os.path.join(OutPath,'genomeSize'),'w') as o:
        for key in sorted(dict_genome.keys()):
            o.writelines('\t'.join([key,str(len(dict_genome[key]))])+'\n')
    
    #Replacing Maternal Genome
    log.log(21,'Repalcing Maternal Genome ...')
    
    for c in Snps.keys():
        pos = Snps[c]['pos']
        for index, i in enumerate(pos):
            dict_genome[c][i - 1] = Snps[c]['m_alt'][index]
        
    #Output Maternal Genome
    M_out_file = os.path.join(OutPath,'Maternal/Maternal.fa')
    log.log(21,'Output Maternal Genome %s ...',M_out_file)
    OutputGenome(dict_genome,M_out_file)
    
    #Replacing Paternal Genome
    log.log(21,'Replacing Paternal Genome ...')
    
    for c in Snps.keys():
        pos = Snps[c]['pos']
        for index, i in enumerate(pos):
            dict_genome[c][i - 1] = Snps[c]['p_alt'][index]
    
    #Output Paternal Genome
    P_out_file = os.path.join(OutPath,'Paternal/Paternal.fa')
    log.log(21,'Output Paternal Genome %s ...',P_out_file)
    OutputGenome(dict_genome,P_out_file)
    
    del dict_genome
    del Snps
    
    #Building Maternal Index and Fragment file
    M_Out_Path = os.path.split(M_out_file)[0]
    enzymeFind(M_out_file,enzyme,M_Out_Path)
    buildIndex(M_out_file,M_Out_Path, threads)
    
    #Building Paternal Index and Fragment file
    P_Out_Path = os.path.split(P_out_file)[0]
    enzymeFind(P_out_file,enzyme,P_Out_Path)
    buildIndex(P_out_file,P_Out_Path, threads)
    


# -*- coding: utf-8 -*-
"""
Created on Thu Mar 01 10:06:52 2018

@author: han-luo
"""

import sys, HiCHap, os, glob, subprocess
from setuptools import  setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if (sys.version_info.major != 2) or (sys.version_info.minor < 6):
    print 'PYTHON VERSION MUST BE 2.6 or 2.7. YOU ARE CURRENTLY USING PYTHON ' + sys.version
    sys.exit(2)

# Guarantee Unix Format
for src in glob.glob('scripts/*'):
    text = open(src, 'r').read().replace('\r\n', '\n')
    open(src, 'w').write(text)
    
setup(name = 'HiCHap',
      version = HiCHap.__version__,
      author = HiCHap.__author__,
      url = 'https://github.com/Prayforhanluo/HiCHap_master',
      author_email = 'hluo_lc@outlook.com',
      description = 'A Library of Hi-C data processing, bias correction and structural analysis for phased haplotype',
      long_description = read('README.rst'),
      long_description_content_type = 'text/x-rst',
      scripts = glob.glob('scripts/*'),
      packages = find_packages(),
      classifiers = [
        'Programming Language :: Python',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)']
      )
      

print "Fix the pkg_resources.ResolutionError::"

execute = subprocess.Popen('which hichap', shell = True, stdout=subprocess.PIPE)

path = execute.stdout.read().strip()
print "Executable file path  : %s " % path
print "Copying build/scripts-2.7/hichap -> %s" %path

subprocess.call('cp ./build/scripts-2.7/hichap %s' % path, shell = True)

print "Install Done!"
